//C++ include
#include <fstream>
#include <map>
#include <set>
#include <string>
#include <vector>
//C includes
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
//Samtools
#include "bam.h"
#include "shortfuse.h"

/* Some globals. */
std::vector<int> refseq_isoform_lengths;
int refseq_isoform_count, N, M;
std::set<std::string> concordant_read_names;
std::set<int> concordant_read_indices;
std::vector<int> fusion_point_rows;
std::vector<int> fusion_point_cols;
std::vector<int> fusion_isoform_lengths;
std::vector<std::string> isoform_names;
std::vector<unsigned int> M_rows;
std::vector<unsigned int> M_cols;
std::vector<double> M_values;
std::vector<double> read_sums;
std::vector<double> theta;
std::vector<double> new_theta;
std::vector<double> distance_probabilities;
std::vector<unsigned int> header1_lengths;
std::vector<unsigned int> fusion_lengths;

double M_value(bam1_t* b1, bam1_t* b2, bool fusion)
{
    double ret_val = 1;
    int m_distance = abs(b1->core.pos - b2->core.pos);
    if(m_distance < 25000)
    {
        ret_val *= distance_probabilities[m_distance];
    }
    else
    {
        return 0;
    }

    int iso_len;
    if(fusion)
    {
        iso_len = fusion_isoform_lengths[b1->core.tid];
    }
    else
    {
        iso_len = refseq_isoform_lengths[b1->core.tid];
    }

    ret_val /= (double)iso_len;

    bam1_t* mcur;
    mcur = bam_init1();

    int i,j;
    for(i=0; i<2; i++)
    {
        if(i==0)
        {
            bam_copy1(mcur, b1);
        }
        else
        {
            bam_copy1(mcur, b2);
        }

        double Pcur = 1;

        /* First get the mismatch indices from the MD tag */
        std::set<short int> mismatches;
        get_mismatch_indices(mismatches, mcur);
        uint8_t* quality = bam1_qual(mcur); /* This is not offset by 33 */
        for(j=0; j<mcur->core.l_qseq; j++)
        {
            double error_probability = pow( (double)10, (double)quality[j]/-10 );
            if(error_probability > .7) error_probability = .7; /* Things get screwy if Q==0 */
            if(mismatches.find(j) == mismatches.end()) /*It's a match */
            {
                Pcur *= 1 - error_probability;
            }
            else
            {
                Pcur *= error_probability / (double)3;
            }
        }
        ret_val *= Pcur;
    }
//    fprintf(stderr, "M_value is returning %e\n", ret_val);
    bam_destroy1(mcur);
    return ret_val;
}

int covers_fusion_point(bam1_t* b)
{
    int start_pos = b->core.pos < b->core.mpos ? b->core.pos : b->core.mpos;
    int end_pos = b->core.pos >= b->core.mpos ? b->core.pos : b->core.mpos;
    end_pos += b->core.l_qseq;

    std::string fusion_transcript = isoform_names[b->core.tid + refseq_isoform_count];
    size_t last_underscore = fusion_transcript.find_last_of("_");
    size_t second_to_last = fusion_transcript.substr(0, last_underscore).find_last_of("_");
    int fusion_point = atoi(fusion_transcript.substr(second_to_last+1, last_underscore).c_str());
    int fuzz = 8;

    int ret_val;
    if(fusion_point > start_pos + fuzz && fusion_point < end_pos - fuzz)
    {
        ret_val = 1;
    }
    else
    {
        ret_val = 0;
    }
    return ret_val;
}

void get_isoform_maps(bamFile refseq1, bamFile refseq2, bamFile fusions)
{
    bam_header_t* refseq1_header;
    bam_header_t* refseq2_header;
    bam_header_t* fusions_header;

    /*Initialize all of the headers*/
    refseq1_header = bam_header_init();
    refseq1_header = bam_header_read(refseq1);
    refseq2_header = bam_header_init();
    refseq2_header = bam_header_read(refseq2);
    fusions_header = bam_header_init();
    fusions_header = bam_header_read(fusions);

    /*First map isoform strings to lengths in columns*/
    int i;
    for(i=0; i<refseq1_header->n_targets; i++)
    {
        std::string isoform_string = std::string(refseq1_header->target_name[i]);
        int first_pipe = isoform_string.find_first_of("|");
        std::string transcript = isoform_string.substr(0, first_pipe);
        isoform_names.push_back(transcript);
        refseq_isoform_lengths.push_back(refseq1_header->target_len[i]);
    }
    refseq_isoform_count = i;

    /*Now add fusion isoforms*/
    for(i=0; i<fusions_header->n_targets; i++)
    {
        std::string isoform_string = std::string(fusions_header->target_name[i]);
        int first_pipe = isoform_string.find_first_of("|");
        std::string transcript = isoform_string.substr(0, first_pipe);
        isoform_names.push_back(transcript);
        /* We also need to update the map from refseq transcripts to fusions.
         * transcript for refseq mappings is like N[RM]_\d+
         * fusion transcripts are like F_N[MR]_\d+\.N[MR]_\d+_\d+_\d+ */
        fusion_isoform_lengths.push_back(fusions_header->target_len[i]);
    }

    M = refseq_isoform_count + i;

    bam_header_destroy(refseq1_header);
    bam_header_destroy(refseq2_header);
    bam_header_destroy(fusions_header);
}

void initialize_mapping_matrix(bamFile refseq1, bamFile refseq2, bamFile fusions)
{
    /* Note that all bamFiles should have read through their headers at this point,
     * and they should be sorted by read. */

    bam1_t** refseq1_mappings = (bam1_t**)malloc(500 * sizeof(bam1_t*));
    bam1_t** refseq2_mappings = (bam1_t**)malloc(500 * sizeof(bam1_t*));
    int i,j;
    for(i=0; i<500; i++)
    {
        refseq1_mappings[i] = bam_init1();
        refseq2_mappings[i] = bam_init1();
    }

    bam1_t* leftovers[2];
    leftovers[0] = bam_init1();
    leftovers[1] = bam_init1();
    bam1_t* fusion_leftovers[2];
    fusion_leftovers[0] = bam_init1();
    fusion_leftovers[1] = bam_init1();
    bam_read1(refseq1, leftovers[0]);
    bam_read1(refseq2, leftovers[1]);
    bam_read1(fusions, fusion_leftovers[0]);
    bam_read1(fusions, fusion_leftovers[1]);

    bam1_t* current_record;
    current_record = bam_init1();
    bam1_t* current_record2;
    current_record2 = bam_init1();

    int r1_mapping_count, r2_mapping_count;

    int read_count = 0;
    int read_fail = 0;
    char current_fus_qname[256];

    while(!read_fail)
    {
        int read_mapped_to_refseq = 0;
        /* First get all the mappings from the first read from refseq1 and refseq2 */
        r1_mapping_count = 1;
        bam_copy1(refseq1_mappings[0], leftovers[0]);
        while(1)
        {
            int read_bytes = bam_read1(refseq1, current_record);
            if(read_bytes <= 0)
            {
                read_fail = 1 ;
                break;
            }
            if(strcmp(bam1_qname(current_record), bam1_qname(refseq1_mappings[0])) == 0)
            {
                bam_copy1(refseq1_mappings[r1_mapping_count], current_record);
                r1_mapping_count++;
            }
            else
            {
                bam_copy1(leftovers[0], current_record);
                break;
            }
        }

        r2_mapping_count = 1;
        bam_copy1(refseq2_mappings[0], leftovers[1]);
        while(1)
        {
            int read_bytes = bam_read1(refseq2, current_record);
            if(read_bytes <= 0)
            {
                read_fail = 1;
                break;
            }
            if(strcmp(bam1_qname(current_record), bam1_qname(refseq2_mappings[0])) == 0)
            {
                bam_copy1(refseq2_mappings[r2_mapping_count], current_record);
                r2_mapping_count++;
            }
            else
            {
                bam_copy1(leftovers[1], current_record);
                break;
            }
        }

        int refseq_fail = 0;
        /* If either end is unmapped, continue */
        if( (refseq1_mappings[0]->core.flag & BAM_FUNMAP) ||
            (refseq2_mappings[0]->core.flag & BAM_FUNMAP) )
            refseq_fail = 1;

        /* Now see if there are good paired mappings between the reads */
        if(!refseq_fail)
        {
            for(i=0; i<r1_mapping_count; i++)
            {
                for(j=0; j<r2_mapping_count; j++)
                {
                    if( (refseq1_mappings[i]->core.tid == refseq2_mappings[j]->core.tid) && /* Same transcript */
                        (abs(refseq1_mappings[i]->core.pos - refseq2_mappings[j]->core.pos) < 2000) && /*Within 2000 nts*/
                        ((refseq1_mappings[i]->core.flag & BAM_FREVERSE) !=
                         (refseq2_mappings[j]->core.flag & BAM_FREVERSE)) && /*In opposite orientations*/
                        (((refseq1_mappings[i]->core.flag & BAM_FREVERSE) && (refseq1_mappings[i]->core.pos > refseq2_mappings[j]->core.pos)) ||
                         ((refseq2_mappings[j]->core.flag & BAM_FREVERSE) && (refseq2_mappings[j]->core.pos > refseq1_mappings[i]->core.pos)))
                         /* The --fr criterion from Bowtie. */)
                    {
                        read_mapped_to_refseq = 1;
                        M_rows.push_back(read_count);
                        M_cols.push_back(refseq1_mappings[i]->core.tid);
                        M_values.push_back(M_value(refseq1_mappings[i], refseq2_mappings[j], 0));
                    }
                }
            }
        }

        /* Now move on to the fusion mappings file.
         * This is a paired mapping BAM file */
        strcpy(current_fus_qname, bam1_qname(fusion_leftovers[0]));
        int read_mapped_to_fusion = 0;
        if(fusion_leftovers[0]->core.flag & BAM_FPROPER_PAIR)
        {
            M_rows.push_back(read_count);
            M_cols.push_back(fusion_leftovers[0]->core.tid + refseq_isoform_count);
            M_values.push_back(M_value(fusion_leftovers[0], fusion_leftovers[1], 1));
            if(covers_fusion_point(fusion_leftovers[0]))
            {
                fusion_point_rows.push_back(read_count);
                fusion_point_cols.push_back(fusion_leftovers[0]->core.tid + refseq_isoform_count);
            }
            read_mapped_to_fusion = 1;
        }
        while(1)
        {
            int read_bytes = bam_read1(fusions, current_record);
            if(read_bytes <= 0){
                read_fail = 1;
                break;
            }
            bam_read1(fusions, current_record2);

            char* current_qname = bam1_qname(current_record);

            if(strcmp(current_qname, bam1_qname(fusion_leftovers[0])) == 0)
            {
                M_rows.push_back(read_count);
                M_cols.push_back(current_record->core.tid + refseq_isoform_count);
                M_values.push_back(M_value(current_record, current_record2, 1));
                if(covers_fusion_point(fusion_leftovers[0]))
                {
                    fusion_point_rows.push_back(read_count);
                    fusion_point_cols.push_back(current_record->core.tid + refseq_isoform_count);
                }
            }
            else
            {
                bam_copy1(fusion_leftovers[0], current_record);
                bam_copy1(fusion_leftovers[1], current_record2);
                break;
            }
        }

        if(read_mapped_to_refseq || read_mapped_to_fusion)
        {
            if(concordant_read_names.find(std::string(current_fus_qname)) != concordant_read_names.end())
            {
                concordant_read_indices.insert(read_count);
            }

            read_count++;
        }
    }

    fprintf(stderr, "Loaded %d mappings from %d reads into memory.\n", (int)M_rows.size(),
            read_count);

    concordant_read_names.clear();
    N = read_count;

    /* Now initialize theta and read count vectors. */
    for(i=0; i<M; i++)
    {
        theta.push_back((double)1/M);
        new_theta.push_back(0);
    }

    for(i=0; i<N; i++)
    {
        read_sums.push_back(0);
    }


    for(i=0; i<500; i++)
    {
        bam_destroy1(refseq1_mappings[i]);
        bam_destroy1(refseq2_mappings[i]);
    }
    free(refseq1_mappings);
    free(refseq2_mappings);
}

void get_distance_probabilities(char* distprobfile)
{
    std::ifstream dpstream(distprobfile);
    std::string nextint, nextprob;

    while(dpstream >> nextint >> nextprob)
    {
        distance_probabilities.push_back(atof(nextprob.c_str()));
    }
    dpstream.close();
}

void EM_update()
{
    int i;

    /* Set read_sums to zero */
    for(i=0; i<N; i++)
    {
        read_sums[i] = 0;
    }

    /* Get the read_sums vector.
     * It is M dot theta. */
    for(i=0; i<(int)M_values.size(); i++)
    {
        int read = M_rows[i];
        int isoform = M_cols[i];
        double Mval = M_values[i];
        read_sums[read] += Mval * theta[isoform];
    }

    /* Now set new theta to zero */
    for(i=0; i<M; i++)
    {
        new_theta[i] = 0;
    }

    /* Get the updated values for new theta */
    for(i=0; i<(int)M_values.size(); i++)
    {
        int read = M_rows[i];
        int isoform = M_cols[i];
        double Mval = M_values[i];
        new_theta[isoform] += Mval * theta[isoform] / read_sums[read] / N;
    }

    /* Copy new_theta to theta */
    theta.assign(new_theta.begin(), new_theta.end());
}

double log_likelihood()
{
    int i;
    double score = 0;

    /* Set read_sums to zero */
    for(i=0; i<N; i++)
    {
        read_sums[i] = 0;
    }

    /* Get the read_sums vector.
     * It is M dot theta. */
    for(i=0; i<(int)M_values.size(); i++)
    {
        int read = M_rows[i];
        int isoform = M_cols[i];
        double Mval = M_values[i];
        read_sums[read] += Mval * theta[isoform];
    }

    for(i=0; i<N; i++)
    {
        if(read_sums[i] > 0) score += log(read_sums[i]);
    }
    return score;
}

void get_concordant_reads(char* concordant_reads_file)
{
    std::ifstream creads(concordant_reads_file);
    std::string cread;

    while(creads >> cread)
    {
        concordant_read_names.insert(cread);
    }
    creads.close();
}

void get_fusion_spanning_counts()
{
    int i, j;
    unsigned int current_row = 0;
    std::set<unsigned int> fusion_rows;
    unsigned int min_col = M + 1;

    for(i=0; i<(int)M_values.size(); i++)
    {
        if(M_rows[i] == current_row)
        {
            if(M_cols[i] < min_col)
            {
                min_col = M_cols[i];
            }
        }
        else
        {
            if(min_col >= (unsigned int)refseq_isoform_count &&
                    concordant_read_indices.find(current_row) == concordant_read_indices.end())
                fusion_rows.insert(current_row);
            current_row = M_rows[i];
            min_col = M_cols[i];
        }
    }

    fprintf(stderr, "Found %d fusion only reads.\n", (int)fusion_rows.size());
    if(min_col >= (unsigned int)refseq_isoform_count &&
            concordant_read_indices.find(current_row) == concordant_read_indices.end())
        fusion_rows.insert(current_row);

    /* Now continue like an EM step
     * Set read_sums to zero */
    for(i=0; i<N; i++)
    {
        read_sums[i] = 0;
    }

    /* And calculate read sums on fusion-only reads. */
    for(i=0; i<(int)M_values.size(); i++)
    {
        int read = M_rows[i];
        if(fusion_rows.find(read) == fusion_rows.end()) continue; /* We only calculate for fusion-only reads */
        int isoform = M_cols[i];
        double Mval = M_values[i];
        read_sums[read] += Mval * theta[isoform];
    }

    /* Now set new theta to zero */
    for(i=0; i<M; i++)
    {
        new_theta[i] = 0;
    }

    /* Get the updated values for new theta */
    for(i=0; i<(int)M_values.size(); i++)
    {
        int read = M_rows[i];
        if(fusion_rows.find(read) == fusion_rows.end()) continue;
        int isoform = M_cols[i];
        int covers = 0;

        for(j=0; j<(int)fusion_point_rows.size(); j++)
        {
            if(read == fusion_point_rows[j] && isoform == fusion_point_cols[j])
            {
               covers = 1;
               break;
            }
        }
        if(!covers) continue;
        double Mval = M_values[i];
        new_theta[isoform] += Mval * theta[isoform] / read_sums[read];
    }
}

int main(int argc, char* argv[])
{
    bamFile refseq1, refseq2, fusions;
    if(argc != 6)
    {
        fprintf(stderr, "EM <refseq.1.bam> <refseq.2.bam> <fusions.bam> <in.distprob> <concordant.reads>\n");
        return 1;
    }
    refseq1 = bam_open(argv[1], "r");
    refseq2 = bam_open(argv[2], "r");
    fusions = bam_open(argv[3], "r");

    get_distance_probabilities(argv[4]);
    get_concordant_reads(argv[5]);
    get_isoform_maps(refseq1, refseq2, fusions);
    initialize_mapping_matrix(refseq1, refseq2, fusions);
    double log_diff = 1;
    int iter_counter = 0;
    while(log_diff > 1e-5)
    {
        iter_counter++;
        double old_ll = log_likelihood();
        EM_update();
        double new_ll = log_likelihood();
        log_diff = fabs(old_ll - new_ll)/fabs(old_ll);
        fprintf(stderr, "Completed EM iteration %d.\n", iter_counter);
        fprintf(stderr, "Old log likelihood was %f.\n", old_ll);
        fprintf(stderr, "New log likelihood was %f.\n", new_ll);
        fprintf(stderr, "Difference is %e.\n", log_diff);
    }

    get_fusion_spanning_counts();

    int i;
    for(i=0; i<M; i++)
    {
        if(new_theta[i] > 0)
        {
            printf("%s\t%3.5f\n", isoform_names[i].c_str(), new_theta[i]);
        }
    }

    return 0;
}

