import getopt
import logging
import os
import re
import subprocess
import sys

bowtie_dir = os.path.expanduser('~/bowtie-0.12.7')
shortfuse_dir = os.path.expanduser('~/ShortFuse')
samtools_dir = os.path.expanduser('~/samtools-0.1.16')
ref_dir = os.path.expanduser('~/ShortFuse/ref')
refseq_bowtie_index = os.path.join(ref_dir, 'refseq_transcripts')
transcripts_plus_genome_index = os.path.join(ref_dir, 'transcripts_plus_genome')
working_dir = os.curdir
processes = 1
python_bin = 'python'
initial_bowtie_flags = '-l 35 -e 150 -n 2 -a -m 150 -S'
discord_bowtie_flags = '-l 22 -e 350 -n 3 -y -a -m 5000'
pe_bowtie_flags = '-a -X 2000 -m 1500 --chunkmbs 1024 -S'
overwrite_existing = "True"

class Pipeline(object):
    
    def __init__(self, fastq1, fastq2):
        
        #Parameters
        self.bowtie_dir = None
        self.shortfuse_dir = None
        self.samtools_dir = None
        self.working_dir = None
        self.processes = None
        self.initial_bowtie_flags = None
        self.discord_bowtie_flags = None
        self.pe_bowtie_flags = None
        self.python_bin = None
        self.refseq_index = None
        self.transcripts_plus_genome_index = None
        self.overwrite_existing = None
        
        #Files created during run
        self.int_files = {}
        self.int_files['fastq1'] = fastq1
        self.int_files['fastq2'] = fastq2
        
    def parse_conf_file(self, conf_file):
        conf_options = ['bowtie_dir', 'shortfuse_dir', 
                        'working_dir', 'ref_dir', 'processes', 'python_bin',
                        'initial_bowtie_flags', 'discord_bowtie_flags',
                        'pe_bowtie_flags', 'overwrite_existing']
                                                
        set_options = []
        for line in file(conf_file):
            l_stripped = line.strip()
            if len(l_stripped) == 0 or l_stripped[0] == '#':
                continue
            l_split = re.split('\s+', l_stripped)
            opt = l_split[0]
            arg = l_split[1]
            if opt not in conf_options:
                print "Ignoring option %s in conf file." % opt
            else:
                print "Setting %s to %s." % (opt, arg)
                arg_u = os.path.expanduser(arg)
                exec('self.' + opt + '="' + arg_u + '"')
                set_options.append(opt)
               
        for opt in conf_options:
            if opt not in set_options:
                print "Using default for %s:%s." % (opt, eval(opt))
                exec('self.' + opt + '="' + eval(opt) + '"')
        
        self.samtools_dir = os.path.join(self.shortfuse_dir, 'samtools')
        self.refseq_index = os.path.join(self.ref_dir, 'refseq_transcripts')
        self.transcripts_plus_genome_index = os.path.join(self.ref_dir, 
                                                    'transcripts_plus_genome')
        

    def run_step(self, step_name, cmd, output_files):
        if eval('self.overwrite_existing'):
            if all([os.path.exists(k) for k in output_files]):
                logging.info("Skipping %s because output files already exist.",
                             step_name)
                return
        t1 = logging.time.time()
        logging.debug(cmd)
        p = subprocess.Popen(cmd, shell=True, stdout = subprocess.PIPE,
                         stderr = subprocess.PIPE)
        stdout, stderr = p.communicate()
        p.poll()
        logging.info(stdout)
        logging.debug(stderr)
        if p.returncode != 0:
            logging.error('%s failed.', step_name)
            sys.exit(p.returncode)
        else:
            t2 = logging.time.time()
            logging.info('%s completed in %dm %ds.', step_name, (t2-t1)//60,
                         (t2-t1)%60)
            
    def map_as_single_reads(self):
        outfile = os.path.join(self.working_dir, 
                        os.path.basename(self.int_files['fastq1']) + '.bam')
        cmd = ' '.join([os.path.join(self.bowtie_dir, 'bowtie'), 
                        self.initial_bowtie_flags, '-p', str(self.processes),
                        os.path.join(self.ref_dir, 'refseq_transcripts'), 
                        self.int_files['fastq1'], '|', 
                        os.path.join(self.samtools_dir, 'samtools'),
                        'view', '-Sb -', '>', outfile])
        self.run_step('Map ' + self.int_files['fastq1'] + ' as single',
                      cmd, [outfile]) 
        self.int_files['bam1'] = outfile
        
        outfile = os.path.join(self.working_dir, 
                        os.path.basename(self.int_files['fastq2']) + '.bam')
        cmd = ' '.join([os.path.join(self.bowtie_dir, 'bowtie'), 
                        self.initial_bowtie_flags, '-p', str(self.processes),
                        os.path.join(self.ref_dir, 'refseq_transcripts'), 
                        self.int_files['fastq2'], '|', 
                        os.path.join(self.samtools_dir, 'samtools'),
                        'view', '-Sb -', '>', outfile])
        self.run_step('Map ' + self.int_files['fastq2'] + ' as single',
                      cmd, [outfile]) 
        self.int_files['bam2'] = outfile
    
    def sort_single_bam_files(self):
        outroot = '.'.join(self.int_files['bam1'].split('.')[:-1]) + '.sorted'
        outfile = outroot + '.bam'
        cmd = ' '.join([os.path.join(self.samtools_dir, 'samtools'),
                        'sort -n', self.int_files['bam1'],
                        outroot])
        self.run_step("Sort " + self.int_files['bam1'],
                      cmd, [outfile])
        self.int_files['bam1sorted'] = outfile
          
        outroot = '.'.join(self.int_files['bam2'].split('.')[:-1]) + '.sorted'
        outfile = outroot + '.bam'
        cmd = ' '.join([os.path.join(self.samtools_dir, 'samtools'),
                        'sort -n', self.int_files['bam2'],
                        outroot])
        self.run_step("Sort " + self.int_files['bam2'],
                      cmd, [outfile])
        self.int_files['bam2sorted'] = outfile

    def sift_for_discordant_pairs(self):
        outprefix = '.'.join(self.int_files['bam1sorted'].split('.')[:-2])
        outfiles = []
        outfiles.append(outprefix + '.distprob')
        outfiles.append(outprefix + '.discord')
        outfiles.append(outprefix + '.discord.fastq')
        cmd = ' '.join([os.path.join(self.shortfuse_dir, 'sift'),
                        self.int_files['bam1sorted'],
                        self.int_files['bam2sorted'],
                        outprefix , '|', 'sort -k 2,2 -k 3,3 >',
                        outprefix + '.discord'])
        self.run_step("Sift", cmd, outfiles)
        self.int_files["discord"] = outprefix + ".discord"
        self.int_files["discordfastq"] = outprefix + ".discord.fastq"
        self.int_files["distprob"] = outprefix + '.distprob'
        
    def discordant_bowtie(self):
        outfiles = []
        outfiles.append(os.path.join(self.working_dir, 'concordant.bt'))
        outfiles.append(self.int_files["discord"] + ".bust")
        cmd = ' '.join([os.path.join(self.bowtie_dir, 'bowtie'),
                        self.discord_bowtie_flags, '-p', str(self.processes),
                        '--max', self.int_files["discord"] + '.bust',
                        self.transcripts_plus_genome_index, 
                        self.int_files["discordfastq"],
                        os.path.join(self.working_dir, "concordant.bt")])
        self.run_step("Discordant bowtie", cmd, outfiles)
        self.int_files["concordantbt"] = os.path.join(self.working_dir, 
                                                                "concordant.bt")
        self.int_files["discordbust"] = self.int_files["discord"] + '.bust'
    
    def parse_discordant_bowtie(self):
        outfile = os.path.join(self.working_dir, 'concordant.reads')
        cmd = ' '.join([self.python_bin, 
                        os.path.join(self.shortfuse_dir, 'parseBowtie.py'),
                        self.int_files["concordantbt"],
                        os.path.join(self.ref_dir, "chrom_regions.txt"),
                        os.path.join(self.working_dir, 
                                     self.int_files["discordbust"]),
                        "100", 
                        ">",os.path.join(self.working_dir, "concordant.reads")])
        self.run_step("Parse discordant Bowtie", cmd, [outfile])
        self.int_files["concordantreads"] = os.path.join(self.working_dir, 
                                                            "concordant.reads")
    def extend_reference(self):
        outfile = self.int_files["discord"] + ".exons"
        cmd = ' '.join([self.python_bin,
                       os.path.join(self.shortfuse_dir, "ExtendReference.py"),
                       os.path.join(self.ref_dir, 'refseq_exon_structure.txt'),
                       self.int_files["discord"], 
                       self.int_files["concordantreads"], "|", "sort -k 2 -u",
                       ">", self.int_files["discord"] + '.exons'])
        self.run_step("Extend reference", cmd, [outfile])
        self.int_files["fusionexons"] = self.int_files["discord"] + '.exons'

    def make_fusion_transcripts(self)        :
        outfile = self.int_files["fusionexons"] + '.fa'
        cmd = ' '.join([self.python_bin,
                        os.path.join(self.shortfuse_dir, "MakeFusionTranscripts.py"),
                        os.path.join(self.ref_dir, "exon_seqs_refseq.fa"),
                        self.int_files["fusionexons"], '>',
                        self.int_files["fusionexons"] + ".fa"])
        self.run_step("Make fusion transcripts", cmd, [outfile])
        self.int_files["fusionseqs"] = self.int_files["fusionexons"] + '.fa'

    def build_fusion_index(self):
        outfile = self.int_files["fusionseqs"] + '.1.ebwt'
        cmd = ' '.join([os.path.join(self.bowtie_dir, "bowtie-build"),
                        '-q -o 0', self.int_files["fusionseqs"], 
                        self.int_files["fusionseqs"]])
        self.run_step("Build fusion index", cmd, [outfile])
    
    def map_paired_end(self):
        outfile = os.path.join(self.working_dir, 
                        os.path.basename(self.int_files['fastq1']) + '.pe.bam')
        cmd = ' '.join([os.path.join(self.bowtie_dir, 'bowtie'), 
                        self.pe_bowtie_flags, '-p', str(self.processes),
                        self.int_files["fusionseqs"], 
                        '-1', self.int_files['fastq1'], 
                        '-2', self.int_files['fastq2'], '|', 
                        os.path.join(self.samtools_dir, 'samtools'),
                        'view', '-Sb -', '>', outfile])
        self.run_step('Map as paired', cmd, [outfile]) 
        self.int_files['pebam'] = outfile

    def sort_paired_end(self)       :
        outroot = '.'.join(self.int_files['pebam'].split('.')[:-1]) + '.sorted'
        outfile = outroot + '.bam'
        cmd = ' '.join([os.path.join(self.samtools_dir, 'samtools'),
                        'sort -n', self.int_files['pebam'],
                        outroot])
        self.run_step("Sort paired end mappings", cmd, [outfile])
        self.int_files['pebamsorted'] = outfile
        
    def em(self):
        outfile = os.path.join(self.working_dir, "fusion_counts.txt")
        cmd = ' '.join([os.path.join(self.shortfuse_dir, "EM"),
                        self.int_files["bam1sorted"],
                        self.int_files["bam2sorted"],
                        self.int_files["pebamsorted"],
                        self.int_files["distprob"],
                        self.int_files["concordantreads"], "|",
                        'sort -grk 2,2 >', outfile])
        self.run_step("Get ML fusion counts", cmd, [outfile])
        self.int_files["fusioncounts"] = outfile
   
    def format_output(self):
        outfile = os.path.join(self.working_dir, 'fusion_counts.bedpe')
        cmd = ' '.join([self.python_bin, 
                       os.path.join(self.shortfuse_dir, "format_output.py"),
                       self.int_files["fusioncounts"],
                       os.path.join(self.ref_dir, 'refseq_ref.txt'),
                       self.int_files["fusionexons"], '>', outfile])
        self.run_step("Format output", cmd, [outfile])
        self.int_files["fusionbedpe"] = outfile
                         
                         
usage = """Usage:
  python run_pipeline --conf <conf_file> <fastq1> <fastq2>

"""
     
if __name__ == '__main__':
        
    try:
        opts, args = getopt.getopt(sys.argv[1:], "h", ["help", "conf="])
    except getopt.GetoptError, err:
        print str(err)
        print usage
        sys.exit(1)
    
    conf_file = None
    for o, a in opts:
        if o in ('-h', '--help'):
            print usage
            sys.exit(0)
        elif o == '--conf':
            conf_file = a
        else:
            raise getopt.GetoptError("Unrecognized option: " + o)
            
    pipeline = Pipeline(args[0], args[1])
    
    if conf_file:
        pipeline.parse_conf_file(conf_file)
    try:
        os.makedirs(pipeline.working_dir)
    except OSError:
        pass

    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    logging.basicConfig(filename=os.path.join(pipeline.working_dir, 
                        os.path.basename(args[0]) + \
                        logging.time.strftime("_%d_%b_%Y_%H_%M_%S") + '.log'),
                        level=logging.DEBUG,
                    format='%(asctime)s %(message)s')
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    console.setFormatter(logging.Formatter('%(message)s'))
    logging.getLogger('').addHandler(console)
    
    start_time = logging.time.time()
    
    pipeline.map_as_single_reads()
    pipeline.sort_single_bam_files()
    pipeline.sift_for_discordant_pairs()
    pipeline.discordant_bowtie()
    pipeline.parse_discordant_bowtie()
    pipeline.extend_reference()
    pipeline.make_fusion_transcripts()
    pipeline.build_fusion_index()
    pipeline.map_paired_end()
    pipeline.sort_paired_end()
    pipeline.em()
    pipeline.format_output()

    end_time = logging.time.time()
    
    logging.info("Overall time for ShortFuse run: %dh %dm %ds", 
                 (end_time-start_time)//3600,
                 ((end_time-start_time)%3600)//60,
                 (end_time-start_time)%60)
    
                         